
(* Cand.sml - to do with candidates - converting name to number, etc *)

structure Cand = struct 

fun mkcc str = implode (List.filter Char.isAlpha (explode str)) ;

(* read electorate file *)
fun readElec fname = 
  let open TextIO ;
    val instr = openIn fname ;
    val str = inputAll instr ;
    val lines = String.fields (fn ch => ch = #"\n") str ;
    fun isBl line = List.all Char.isSpace (explode line) ;
    val groups = Gen.lfields (not o isBl) lines ;
    val pcs = map (fn op :: ht => ht) groups ;
  in pcs end ;

fun snd (a, b) = b ;

(* getccs : ('a * string list) list ->
  string vector * (int * string) btree * (string -> int) * (int -> string)
  put candidate names and numbers into vector (for looking up by number
  and binary search tree (for looking up by name);
  also returns conversion functions *)
fun getccs pcs = 
  let
    val cands = List.concat (map snd pcs) ;
    val ccs = map mkcc cands ;
    val candNames = Vector.fromList ("" :: ccs) ;
    val ncs = Types.numl 1 ccs ;
    val sncs = Listsort.sort (Gen.trComp snd String.compare) ncs ;
    val (cnst, []) = Gen.mkBt (length sncs) sncs ;
    fun candNum name = 
      case Gen.findBt (Gen.trComp snd String.compare) (0, name) cnst of 
	  NONE => raise Fail ("candidate name " ^ name ^ " not found")
	| SOME (num, name) => num ;
    fun candName 0 = raise Fail "no candidate 0"
      | candName n = Vector.sub (candNames, n) 
        handle Subscript => raise Fail "candidate number negative or too high" ;
  in (candNames, cnst, candNum, candName) end ;

(* s2clvote : (int -> string) -> string -> string
  cl2svote : (string -> int) -> string -> string
  convert vote between string and vote using cand codes *)

fun s2clvote candName votestr = 
  let val il = RPV.s2ilvote votestr ;
    val names = map candName il ;
    open Msp ; infix && ;
    val line = prsep ($ ",") $ names && $ "\n" ;
  in flatten line end ;

fun cl2svote candNum str = 
  let val names = 
      String.tokens (fn ch => ch = #"," orelse Char.isSpace ch) str ;
    val nums = map candNum names ;
  in case nums of [] => NONE
      | nums => SOME (RPV.il2svote nums) end ;

(* votes stored in string format, but read and write to file
  in cand code format *)

(* prints2clvotes : (int -> string) -> string list -> string -> unit *)
fun prints2clvotes candName votes file = 
  let val ostr = TextIO.openOut file ;
    fun prv vote = TextIO.output (ostr, s2clvote candName vote) ;
    val () = app prv votes ;
    val () = TextIO.closeOut ostr ;
  in () end ;

(* readcl2svotes : (string -> int) -> int -> string -> int * string list 
  given file of votes in candidate list format, 
  read and turn into string format, ignoring blank lines (informal votes) ;
  read up to nv votes, also return number of formal votes read *)
(* note - input line of "\n" is blank (empty) line;
  in case cl2svote candNum str of NONE this indicates a 
  line containing only white space;
  both treated as valid input representing an informal vote *)
fun readcl2svotes candNum nballots file = 
  let open TextIO ;
    val instr = openIn file ;
    fun rvs 0 = (0, []) 
      | rvs n_to_read = 
	case inputLine instr of "" => (0, [])
	  | "\n" => rvs (n_to_read - 1) 
	  | str => 
	    case cl2svote candNum str of NONE => rvs (n_to_read - 1) 
	      | SOME vote => 
		let val (nfv, vs) = rvs (n_to_read - 1) ;
		in (nfv+1, vote :: vs) end ;
  in rvs nballots before closeIn instr end ;

(* this version is tail recursive, and works where the other one fails
  in PolyML due to lack of swap space *)
fun readcl2svotes candNum nballots file = 
  let open TextIO ;
    val instr = openIn file ;
    fun rvs 0 sf = sf
      | rvs n_to_read (sf as (nsf, vsf)) = 
	case inputLine instr of "" => sf
	  | "\n" => rvs (n_to_read - 1) sf (* blank line - ignored *)
	  | str => 
	    case cl2svote candNum str of NONE => rvs (n_to_read - 1) sf 
	      | SOME vote => rvs (n_to_read - 1) (nsf + 1, vote :: vsf) ;
    val (nfv, rev_vs) = (rvs nballots (0, [])) ;
    val () = closeIn instr ;
  in (nfv, rev rev_vs) end ;

(* readclv : string -> string -> int -> int * string list
  given electorate file and file of votes in format like
  CarolineLECOUTEUR,JimCOATES,RolandMANDERSON,JaneERREY\n,
  get list of votes in string format,
  read up to nv votes, return number of formal votes read *)
fun readclv elecfile votefile nv = 
  let 
    val pcs = readElec elecfile ;
    val (candNames, cnst, candNum, candName) = getccs pcs ;
  in readcl2svotes candNum nv votefile end ;

end ; (* structure Cand *)

(* 
open Cand ;
val fname = "Fraser" ; 
val pcs = readElec fname ;
val (candNames, cnst, candNum, candName) = getccs pcs ;
map candNum ccs ;
map candName it ;

prints2clvotes candName votes "/scratch/votefile.cl" ;
*)


