
(* 
app load ["Types", "Listsort"] ;
*)
structure Tie = struct 
open Types ;

(* selcts : ('a -> cand) -> 'a list -> (cand * int) list -> ('a * int) list
  where map arg1 arg2 is a list of candidates, and
  arg3 is a list of candidates and vote totals (both lists in order),
  return list of each 'a-value and vote total *)
fun selcts f [] cvl = []
  | selcts f (ct :: cts) (cvs as (c, v) :: cvs') = 
    case Int.compare (f ct, c) of EQUAL => (ct, v) :: selcts f cts cvs'
      | GREATER => selcts f (ct :: cts) cvs'
      | LESS => (ct, 0) :: selcts f cts cvs ;

(* allDiff : ('a * int) list -> 'a list option
  sorts list by second component, returns list of first components
  if all different *)
fun allDiff cvs = 
  let fun comp ((c, v), (c', v')) = Int.compare (v, v') ;
    val cvsort = Listsort.sort comp cvs ;
    fun alld (cv1 :: cv2 :: cvs) = 
	comp (cv1, cv2) <> EQUAL andalso alld (cv2 :: cvs)
      | alld [cv] = true
      | alld [] = raise Fail "allDiff : empty" ;
  in if alld cvsort then SOME (map #1 cvsort) else NONE end ;

(* contempsDiff : ('a -> cand) -> 'a list -> (count * results) list ->
  (count * ('a * int) list) option
  where map f cts gives a list of candidates,
  go back to find count where all these candidates have different totals,
  return number of that count and list of ('a, total vote) pairs;
  arg3 is list of previous counts and their results;
  returns NONE if no count on which candidate totals all differ,
  in which case Commissioner's lot is needed *)
fun contempsDiff f cts [] = NONE
  | contempsDiff f cts ((count, Cres {tot = (_, cvl), ...}) :: res) = 
    (* (selcts f cts cvl) is list of 'a-values and vote totals for the
      candidates at the most recent count included in arg3;
      if these are all different, then return these, with the count;
      otherwise, go back to earlier counts *)
    (case allDiff (selcts f cts cvl) of SOME cvsort => SOME (count, cvsort)
      | NONE => contempsDiff f cts res) ;

fun commLot' (info as Info {commLot, ...}) cnt cands = 
  let val pos = commLot cnt cands ; (* result, as position in list *) 
    val succ = List.nth (cands, pos) ; (* successful candidate *)
  in Gen.recordEvent info (CommLot (cands, succ)) ; pos end ;

fun commLotExcl (info as Info {commLot, ...}) cnt cts = 
  let val pos = commLot' info cnt cts ;
  in List.nth (cts, pos) end ;

(* get and delete nth (counting from 0) of list *)
fun get'delete 0 (x :: xs) = (x, xs)
  | get'delete n (x :: xs) = 
    let val (y, ys) = get'delete (n-1) xs ;
    in (y, x :: ys) end
  | get'delete _ [] = raise Fail "get'delete: position not in list" ;

fun commLotSurp (info as Info {commLot, ...}) cnt cts = 
  let
    fun order [c1] = [c1]
      | order cnds = 
	let val pos = commLot' info cnt (map ue1 cnds) ;
	  val (chosen, rest) = get'delete pos cnds ;
	in chosen :: order rest end ;
  in order cts end ;

(* chooseExcl : cand list -> info -> int * cand
  excludes candidate per cl 8(2)(a) *)
fun chooseExcl contemps (info as Info {results, count, ...}) = 
  let val () = 
      if (#1 o hd o !) results <> !count then raise Fail "chooseExcl" else () ;
  in case contempsDiff (fn x => x) contemps (tl (!results)) of
	SOME (cnt, c :: _) => (cnt, c) 
      | NONE => (0, commLotExcl info (!count) contemps) end ;

fun fst (a, b) = a ;

(* orderSurp : ('a -> int) -> 'a list -> info -> int * 'a list
  determine the order of contemporary candidates when all have equal
  surplus - clause 7(3)(c) *)
fun orderSurp f contemps (info as Info {results, count, ...}) = 
  let val () = 
      if !count <> 1 andalso (#1 o hd o !) results <> (!count - 1) then 
	raise Fail "orderSurp"
      else () ;
  in case contempsDiff f contemps (!results) of 
	SOME (cnt, cvs) => (cnt, rev cvs) (* sorted in decreasing order *)
      | NONE => (0, commLotSurp info (!count) contemps) end ;

(* tdw : ('a -> bool) -> 'a list -> 'a list * 'a list
  take-drop-while, divides list into two, first part is max prefix
  all satisfying p *)
fun tdw p [] = ([], [])
  | tdw p (x :: xs) = 
    if p x then 
      let val (sps, rest) = tdw p xs in (x :: sps, rest) end 
    else ([], x :: xs) ;

(* collEq : ('b * 'b -> order) -> 'b list -> 'b list list
  groups list into sublists of items which compare equal *) 
fun collEq comp [] = []
  | collEq comp (x :: xs) = 
    let val (eqx, rest) = tdw (fn x' => comp (x, x') = EQUAL) xs ;
    in (x :: eqx) :: collEq comp rest end ;

(* sortElect : elec list -> info -> elec list
  determine order of several candidates with a surplus
  obtained at the same count - clause 7(3)(b)&(c) *)
fun sortElect (elects as _ :: _ :: _) (info as Info {count, events, ...}) = 
    let 
      fun comp (Elec a, Elec b) = 
	Int.compare (#2 b, #2 a) ; (* want decreasing *)
      val sel = Listsort.sort comp elects ; (* list sorted by votes *)
      val selg = collEq comp sel ; (* group those with equal numbers *)
      (* where candidates have same number of votes, resolve
	by looking back - clause 7(3)(c) *)
      fun resolve [c] = [c]
	| resolve [] = raise Fail "sortElect" 
	| resolve (contemps as Elec (_, n, _, ct) :: _) = 
	  let val () = if ct <> !count then raise Fail "sortElect - count"
	      else () ;
	    val (cnt, elecs) = orderSurp ue1 contemps info ;
	    val () = Gen.recordEvent info (SurpTie {count = cnt, 
	      order = map ue1 elecs, votes = n}) ;
	  in elecs end ;
      val selgr = List.concat (map resolve selg) ;
    in selgr end 
  | sortElect elects _ = elects ;

end ; (* structure Tie *)

